@@ -1,6 +1,7 @@
 """Imports of dynamic libraries used for text layout."""
 
 import os
+from contextlib import suppress
 
 import cffi
 
@@ -70,6 +71,12 @@ ffi.cdef('''
         PANGO_WEIGHT_ULTRAHEAVY = 1000
     } PangoWeight;
 
+    typedef enum {
+        PANGO_FONT_MASK_SIZE = 1 << 5,
+        PANGO_FONT_MASK_GRAVITY = 1 << 6,
+        PANGO_FONT_MASK_VARIATIONS = 1 << 7
+    } PangoFontMask;
+
     typedef enum {
         PANGO_STRETCH_ULTRA_CONDENSED,
         PANGO_STRETCH_EXTRA_CONDENSED,
@@ -104,44 +111,6 @@ ffi.cdef('''
        struct GSList *next;
     } GSList;
 
-    typedef struct {
-        const PangoAttrClass *klass;
-        guint start_index;
-        guint end_index;
-    } PangoAttribute;
-
-    typedef struct {
-        PangoLayout *layout;
-        gint         start_index;
-        gint         length;
-        GSList      *runs;
-        guint        is_paragraph_start : 1;
-        guint        resolved_dir : 3;
-    } PangoLayoutLine;
-
-    typedef struct  {
-        int x;
-        int y;
-        int width;
-        int height;
-    } PangoRectangle;
-
-    typedef struct {
-        guint is_line_break: 1;
-        guint is_mandatory_break : 1;
-        guint is_char_break : 1;
-        guint is_white : 1;
-        guint is_cursor_position : 1;
-        guint is_word_start : 1;
-        guint is_word_end : 1;
-        guint is_sentence_boundary : 1;
-        guint is_sentence_start : 1;
-        guint is_sentence_end : 1;
-        guint backspace_deletes_character : 1;
-        guint is_expandable_space : 1;
-        guint is_word_boundary : 1;
-    } PangoLogAttr;
-
     typedef struct {
         void *shape_engine;
         void *lang_engine;
@@ -188,6 +157,49 @@ ffi.cdef('''
         PangoGlyphString *glyphs;
     } PangoGlyphItem;
 
+    typedef struct GSListRuns {
+       PangoGlyphItem    *data;
+       struct GSListRuns *next;
+    } GSListRuns;
+
+    typedef struct {
+        const PangoAttrClass *klass;
+        guint start_index;
+        guint end_index;
+    } PangoAttribute;
+
+    typedef struct {
+        PangoLayout *layout;
+        gint         start_index;
+        gint         length;
+        GSListRuns  *runs;
+        guint        is_paragraph_start : 1;
+        guint        resolved_dir : 3;
+    } PangoLayoutLine;
+
+    typedef struct  {
+        int x;
+        int y;
+        int width;
+        int height;
+    } PangoRectangle;
+
+    typedef struct {
+        guint is_line_break: 1;
+        guint is_mandatory_break : 1;
+        guint is_char_break : 1;
+        guint is_white : 1;
+        guint is_cursor_position : 1;
+        guint is_word_start : 1;
+        guint is_word_end : 1;
+        guint is_sentence_boundary : 1;
+        guint is_sentence_start : 1;
+        guint is_sentence_end : 1;
+        guint backspace_deletes_character : 1;
+        guint is_expandable_space : 1;
+        guint is_word_boundary : 1;
+    } PangoLogAttr;
+
     int pango_version (void);
 
     double pango_units_to_double (int i);
@@ -220,20 +232,31 @@ ffi.cdef('''
     void pango_font_description_free (PangoFontDescription *desc);
     PangoFontDescription * pango_font_description_copy (
         const PangoFontDescription *desc);
+
     void pango_font_description_set_family (
         PangoFontDescription *desc, const char *family);
     void pango_font_description_set_style (
         PangoFontDescription *desc, PangoStyle style);
-    PangoStyle pango_font_description_get_style (
-        const PangoFontDescription *desc);
     void pango_font_description_set_stretch (
         PangoFontDescription *desc, PangoStretch stretch);
     void pango_font_description_set_weight (
         PangoFontDescription *desc, PangoWeight weight);
     void pango_font_description_set_absolute_size (
         PangoFontDescription *desc, double size);
+    void pango_font_description_set_variations (
+        PangoFontDescription* desc, const char* variations);
+
+    PangoStyle pango_font_description_get_style (
+        const PangoFontDescription *desc);
+    const char* pango_font_description_get_variations (
+        const PangoFontDescription* desc);
+    PangoWeight pango_font_description_get_weight (
+        const PangoFontDescription* desc);
     int pango_font_description_get_size (PangoFontDescription *desc);
 
+    void pango_font_description_unset_fields (
+        PangoFontDescription* desc, PangoFontMask to_unset);
+
     int pango_glyph_string_get_width (PangoGlyphString *glyphs);
     char * pango_font_description_to_string (
         const PangoFontDescription *desc);
@@ -241,7 +264,7 @@ ffi.cdef('''
     PangoFontDescription * pango_font_describe (PangoFont *font);
     const char * pango_font_description_get_family (
         const PangoFontDescription *desc);
-    int pango_font_description_hash (const PangoFontDescription *desc);
+    guint pango_font_description_hash (const PangoFontDescription *desc);
 
     PangoContext * pango_context_new ();
     PangoContext * pango_font_map_create_context (PangoFontMap *fontmap);
@@ -249,6 +272,8 @@ ffi.cdef('''
     PangoFontMetrics * pango_context_get_metrics (
         PangoContext *context, const PangoFontDescription *desc,
         PangoLanguage *language);
+    PangoFontMetrics * pango_font_get_metrics (
+        PangoFont *font, PangoLanguage *language);
     void pango_font_metrics_unref (PangoFontMetrics *metrics);
     int pango_font_metrics_get_ascent (PangoFontMetrics *metrics);
     int pango_font_metrics_get_descent (PangoFontMetrics *metrics);
@@ -260,17 +285,13 @@ ffi.cdef('''
         PangoFontMetrics *metrics);
     int pango_font_metrics_get_strikethrough_position (
         PangoFontMetrics *metrics);
-
-    void pango_context_set_round_glyph_positions (
-        PangoContext *context, gboolean round_positions);
-
-    PangoFontMetrics * pango_font_get_metrics (
-        PangoFont *font, PangoLanguage *language);
-
     void pango_font_get_glyph_extents (
         PangoFont *font, PangoGlyph glyph, PangoRectangle *ink_rect,
         PangoRectangle *logical_rect);
 
+    void pango_context_set_round_glyph_positions (
+        PangoContext *context, gboolean round_positions);
+
     PangoAttrList * pango_attr_list_new (void);
     void pango_attr_list_unref (PangoAttrList *list);
     void pango_attr_list_insert (
@@ -377,23 +398,28 @@ ffi.cdef('''
 def _dlopen(ffi, *names):
     """Try various names for the same library, for different platforms."""
     for name in names:
-        try:
+        with suppress(OSError):
             return ffi.dlopen(name)
-        except OSError:
-            pass
     # Re-raise the exception.
+    print(
+        '\n-----\n\n'
+        'WeasyPrint could not import some external libraries. Please '
+        'carefully follow the installation steps before reporting an issue:\n'
+        'https://doc.courtbouillon.org/weasyprint/stable/'
+        'first_steps.html#installation\n'
+        'https://doc.courtbouillon.org/weasyprint/stable/'
+        'first_steps.html#troubleshooting',
+        '\n\n-----\n')  # pragma: no cover
     return ffi.dlopen(names[0])  # pragma: no cover
 
 
-if hasattr(os, 'add_dll_directory'):
+if hasattr(os, 'add_dll_directory'):  # pragma: no cover
     dll_directories = os.getenv(
         'WEASYPRINT_DLL_DIRECTORIES',
         'C:\\Program Files\\GTK3-Runtime Win64\\bin').split(';')
     for dll_directory in dll_directories:
-        try:
+        with suppress((OSError, FileNotFoundError)):
             os.add_dll_directory(dll_directory)
-        except (OSError, FileNotFoundError):
-            pass
 
 gobject = _dlopen(
     ffi, 'gobject-2.0-0', 'gobject-2.0', 'libgobject-2.0-0',
@@ -407,7 +433,7 @@ harfbuzz = _dlopen(
     'libharfbuzz-0.dll')
 fontconfig = _dlopen(
     ffi, 'fontconfig-1', 'fontconfig', 'libfontconfig', 'libfontconfig.so.1',
-    'libfontconfig-1.dylib', 'libfontconfig-1.dll')
+    'libfontconfig.1.dylib', 'libfontconfig-1.dll')
 pangoft2 = _dlopen(
     ffi, 'pangoft2-1.0-0', 'pangoft2-1.0', 'libpangoft2-1.0-0',
     'libpangoft2-1.0.so.0', 'libpangoft2-1.0.dylib', 'libpangoft2-1.0-0.dll')
